//this file is part of Notepad++
//Copyright (C)2007 Don HO <donho@altern.org>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "PluginInterface.h"
//#include <stdio.h>

const char PLUGIN_NAME[] = "Base64";
const int nbFunc = 2;

NppData nppData;
FuncItem funcItem[nbFunc];

void convertToBase64FromAscii();
void convertToAsciiFromBase64();


#define ASCII_BITLEN 8
#define B64_BITLEN 6

char base64Array[65] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
int base64IndexArray[123] =\
{\
-1, -1, -1, -1, -1, -1, -1, -1,\
-1, -1, -1, -1, -1, -1, -1, -1,\
-1, -1, -1, -1, -1, -1, -1, -1,\
-1, -1, -1, -1, -1, -1, -1, -1,\
-1, -1, -1, -1, -1, -1, -1, -1,\
-1, -1, -1, 62, -1, -1, -1, 63,\
52, 53, 54, 55 ,56, 57, 58, 59,\
60, 61, -1, -1, -1, -1, -1, -1,\
-1,  0,  1,  2,  3,  4,  5,  6,\
 7,  8,  9, 10, 11, 12, 13, 14,\
15, 16, 17, 18, 19, 20, 21, 22,\
23, 24, 25, -1, -1, -1, -1 ,-1,\
-1, 26, 27, 28, 29, 30, 31, 32,\
33, 34, 35, 36, 37, 38, 39, 40,\
41, 42, 43, 44, 45, 46, 47, 48,\
49, 50, 51\
};

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  reasonForCall, 
                       LPVOID lpReserved )
{
	switch (reasonForCall)
	{
		case DLL_PROCESS_ATTACH:
		{
			funcItem[0]._pFunc = convertToBase64FromAscii;
			funcItem[1]._pFunc = convertToAsciiFromBase64;

			strcpy(funcItem[0]._itemName, "Base64 Encode");
			strcpy(funcItem[1]._itemName, "Base64 Decode");

			funcItem[0]._init2Check = false;
			funcItem[1]._init2Check = false;

			// If you don't need the shortcut, you have to make it NULL
			funcItem[0]._pShKey = NULL;
			funcItem[1]._pShKey = NULL;
		}
		break;

		case DLL_PROCESS_DETACH:
		break;

		case DLL_THREAD_ATTACH:
		break;

		case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}

extern "C" __declspec(dllexport) void setInfo(NppData notpadPlusData)
{
	nppData = notpadPlusData;
}

extern "C" __declspec(dllexport) const char * getName()
{
	return PLUGIN_NAME;
}

extern "C" __declspec(dllexport) FuncItem * getFuncsArray(int *nbF)
{
	*nbF = nbFunc;
	return funcItem;
}

extern "C" __declspec(dllexport) void beNotified(SCNotification *notifyCode)
{	
}

// Here you can process the Npp Messages 
// I will make the messages accessible little by little, according to the need of plugin development.
// Please let me know if you need to access to some messages :
// http://sourceforge.net/forum/forum.php?forum_id=482781
//
extern "C" __declspec(dllexport) LRESULT messageProc(UINT Message, WPARAM wParam, LPARAM lParam)
{
	return TRUE;
}


HWND getCurrentScintillaHandle()
{
    int currentEdit;
    ::SendMessage(nppData._nppHandle, NPPM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
	return (currentEdit == 0)?nppData._scintillaMainHandle:nppData._scintillaSecondHandle;
};

size_t getSelStartPos(HWND hScitilla)
{
	size_t start = ::SendMessage(hScitilla, SCI_GETSELECTIONSTART, 0, 0);
	size_t end = ::SendMessage(hScitilla, SCI_GETSELECTIONEND, 0, 0);
	return (start < end)?start:end;
}


size_t getBase64LenFromAsciiLen(size_t asciiStrLen)
{
	size_t nbBase64Byte = asciiStrLen * ASCII_BITLEN / B64_BITLEN;
	return nbBase64Byte + (((asciiStrLen * ASCII_BITLEN) % B64_BITLEN)?3:0);
};

size_t getAsciiLenFromBase64Len(size_t base64StrLen)
{
	if (base64StrLen % 4) return 0;
	return  base64StrLen - base64StrLen / 4;
};

void asciiToBase64(char *dest, const char *asciiStr)
{
	size_t asciiStrLen = strlen(asciiStr);
	size_t nbLoop = asciiStrLen / 3;
	size_t nbByteRemain = asciiStrLen % 3;

	size_t i = 0, k = 0;
	for ( ; i < nbLoop ; i++)
	{
		size_t j = i * 3;
		unsigned char p1, p2;

		unsigned char a = asciiStr[j] >> 2;
		dest[k++] = base64Array[a];
		p1 = asciiStr[j] << 6;
		p2 = asciiStr[j+1] >> 2;
		unsigned char b = p1 | p2;
		b >>= 2;
		dest[k++] = base64Array[b];

		p1 = asciiStr[j+1] << 4;
		p2 = asciiStr[j+2] >> 4;
		unsigned char c = p1 | p2;
		c >>= 2;
		dest[k++] = base64Array[c];

		p2 = asciiStr[j+2] << 2;
		unsigned char d = p2 >> 2;
		dest[k++] = base64Array[d];
	}

	if (nbByteRemain == 1)
	{
		size_t j = i * 3;
		unsigned char p;
		dest[k++] = base64Array[asciiStr[j] >> 2];

		p = asciiStr[j] << 6;
		dest[k++] = base64Array[p >> 2];

		dest[k++] = '=';
		dest[k++] = '=';
	}
	else if (nbByteRemain == 2)
	{
		size_t j = i * 3;
		unsigned char p1, p2;

		dest[k++] = base64Array[asciiStr[j] >> 2];

		p1 = asciiStr[j] << 6;
		p2 = asciiStr[j+1] >> 2;
		unsigned char b = p1 | p2;
		dest[k++] = base64Array[b >> 2];

		p1 = asciiStr[j+1] << 4;
		dest[k++] = base64Array[p1 >> 2];

		dest[k++] = '=';
	}
	dest[k] = '\0';
}

bool base64ToAscii(char *dest, const char *base64Str)
{
	size_t base64StrLen = strlen(base64Str);
	size_t nbLoop = base64StrLen / 4;
	if (base64StrLen % 4)
		return false;
	size_t i = 0, k = 0;

	//enum {b64_just, b64_1padded, b64_2padded} padd = b64_just;
	for ( ; i < nbLoop ; i++)
	{
		size_t j = i * 4;
		unsigned char uc0, uc1, uc2, uc3, p0, p1;

		uc0 = base64IndexArray[base64Str[j]];
		uc1 = base64IndexArray[base64Str[j+1]];
		uc2 = base64IndexArray[base64Str[j+2]];
		uc3 = base64IndexArray[base64Str[j+3]];

		if ((uc0 == -1) || (uc1 == -1) || (uc2 == -1) || (uc3 == -1))
			return false;

		if (base64Str[j+2] == '=') // && (uc3 == '=')
		{
			uc2 = uc3 = 0;
			//padd = b64_2padded;
		}
		else if (base64Str[j+3] == '=')
		{
			uc3 = 0;
			//padd = b64_1padded;
		}
		p0 = uc0 << 2;
		p1 = uc1 << 2;
		p1 >>= 6;
		dest[k++] = p0 | p1;

		p0 = uc1 << 4;
		p1 = uc2 << 2;
		p1 >>= 4;
		dest[k++] = p0 | p1;

		p0 = uc2 << 6;
		p1 = uc3;
		dest[k++] = p0 | p1;
	}
	dest[k] = '\0';
	//if (padd == b64_1padded)
	//	dest[k-1] = '\0';
	//else if (padd == b64_2padded)
	//	dest[k-2] = '\0';

	return true;
}

void convertToBase64FromAscii()
{	 
	HWND hCurrScintilla = getCurrentScintillaHandle();
    size_t textLength = ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, 0);
    if (textLength == 0) return;

    char * pAsciiText = new char[textLength];
    
    ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, (LPARAM)pAsciiText);

	size_t b64Len = getBase64LenFromAsciiLen(textLength);
	char * pBase64Text = new char[b64Len + 1];

	asciiToBase64(pBase64Text, pAsciiText);
	size_t start = getSelStartPos(hCurrScintilla);
	::SendMessage(hCurrScintilla, SCI_REPLACESEL, 0, (LPARAM)pBase64Text);
	::SendMessage(hCurrScintilla, SCI_SETSEL, start, start+strlen(pBase64Text));

	delete [] pAsciiText;
	delete [] pBase64Text;
}

void convertToAsciiFromBase64()
{
	HWND hCurrScintilla = getCurrentScintillaHandle();
    size_t textLength = ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, 0);
    if (textLength == 0) return;

	if ((textLength - 1) % 4)
	{
		::MessageBox(nppData._nppHandle, "Length of selected text to be decoded is invalid.\rIt should be mod 4.", "Base64", MB_OK);
		return;
	}

    char * pBase64Text = new char[textLength];
    ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, (LPARAM)pBase64Text);

	size_t asciiLen = getAsciiLenFromBase64Len(textLength - 1);
	char * pAsciiText = new char[asciiLen + 1];

	bool isGood = base64ToAscii(pAsciiText, pBase64Text);
	if (isGood)
	{
		size_t start = getSelStartPos(hCurrScintilla);
		::SendMessage(hCurrScintilla, SCI_REPLACESEL, 0, (LPARAM)pAsciiText);
		::SendMessage(hCurrScintilla, SCI_SETSEL, start, start+strlen(pAsciiText));
	}
	else
		::MessageBox(nppData._nppHandle, "Problem!", "Base64", MB_OK);

	delete [] pAsciiText;
	delete [] pBase64Text;
}

